# flake8: noqa: E501

import unittest
from lxml import etree

from lift_utils import base
from lift_utils import config
from lift_utils import datatypes
from lift_utils import utils

from lift_utils.lexicon import Etymology

from .utils import compare_xml_trees

LIFT_VERSION = '0.15'


class TestAnnotation(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Annotation()
        obj.name = datatypes.Key("annotation-name")
        obj.value = datatypes.Key("annotation-value")
        obj.who = datatypes.Key("Annotation Who")
        obj.when = datatypes.DateTime("0000-00-00")

        xml = f"""
        <annotation name="{obj.name}" value="{obj.value}" who="{obj.who}" when="{obj.when}"/>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestField(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Field()
        obj.date_created = datatypes.DateTime("0000-00-00")
        obj.date_modified = datatypes.DateTime("0000-00-00")
        obj.name = datatypes.Key("field-type")
        obj.annotation_items = [
            base.Annotation(),
            base.Annotation(),
        ]
        obj.trait_items = [
            base.Trait(),
            base.Trait(),
        ]
        obj.form_items = [
            base.Form(),
            base.Form(),
        ]

        xml = f"""
        <field dateCreated="{obj.date_created}" dateModified="{obj.date_modified}" name="{obj.name}">
            <form/>
            <form/>
            <trait/>
            <trait/>
            <annotation/>
            <annotation/>
        </field>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        # compare_xml_trees(obj._to_xml_tree(), xml_tree)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestForm(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Form()
        obj.lang = datatypes.Lang('en')
        obj.text = base.Text()
        obj.annotation_items = [
            base.Annotation(),
            base.Annotation(),
        ]

        xml = f"""
        <form lang="{obj.lang}">
            <text/>
            <annotation/>
            <annotation/>
        </form>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestGloss(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Gloss()
        obj.trait_items = [
            base.Trait(),
            base.Trait(),
        ]

        xml = f"""
        <gloss>
            <trait/>
            <trait/>
        </gloss>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestSpan(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Span()
        obj.lang = datatypes.Lang('en')
        obj.href = datatypes.URL("https://example.com")
        obj.class_ = "TestClass"
        obj.pcdata = datatypes.PCData("test text")
        obj.tail = datatypes.PCData(" a tail.")

        xml = f"""
        <xml>
        <span lang="{obj.lang}" href="{obj.href}" class="{obj.class_}">{obj.pcdata}</span>{obj.tail}</xml>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        xml_tree = xml_tree.getchildren()[0]
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestText(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Text()
        obj.pcdata = datatypes.PCData('Test text')
        obj.span_items = [
            base.Span(),
            base.Span(),
        ]

        xml = f"""
        <text>{obj.pcdata}<span/><span/></text>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestTrait(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.Trait()
        obj.name = datatypes.Key('TraitName')
        obj.value = datatypes.Key('TraitValue')
        obj.id = datatypes.Key('TraitID')
        obj.annotation_items = [
            base.Annotation(),
            base.Annotation(),
        ]

        xml = f"""
        <trait name="{obj.name}" value="{obj.value}" id="{obj.id}">
            <annotation/>
            <annotation/>
        </trait>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestURLRef(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = base.URLRef()
        obj.href = datatypes.URL("https://example.com")
        obj.label = base.Multitext()

        xml = f"""
        <urlref href="{obj.href}">
            <label/>
        </urlref>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )



"""Define the basic datatypes."""

import uuid
from typing import List


class PCData(str):
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class DateTime(str):
    # format (str): YYYY-MM-DDTHH:MM:SSZZZZZZ
    # ZZZZZZ: +/-, H, H, :, M, M (offset from GMT)
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class Key(str):
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class Lang(str):
    # format (str): ISO[-SCRIPT[-x-PRIVATE]]
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class RefId(str):
    # format (HEX GUID): xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)
        else:  # generate new UUID
            return super().__new__(cls, str(uuid.uuid4()))


class URL(str):
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class Prop:
    def __init__(
        self,
        name: str = None,
        required: bool = False,
        prop_type=None,
        item_type=None,
    ):
        self.name = name
        self.required = required
        self.prop_type = prop_type
        self.item_type = item_type


class Props:
    def __init__(
        self,
        lift_version: str = None,
        attributes: List[Prop] = None,
        elements: List[Prop] = None,
    ):
        self.lift_version = lift_version
        self.attributes = attributes
        self.elements = elements

    def add_to(self, prop_group_name, prop_obj):
        prop_group = self.__dict__.get(prop_group_name)
        if prop_obj.name not in (p.name for p in prop_group):
            prop_group.append(prop_obj)



# flake8: noqa

import unittest
from lxml import etree

from .utils import compare_xml_trees

from lift_utils import base
from lift_utils import config
from lift_utils import datatypes
from lift_utils import lexicon
from lift_utils import utils

LIFT_VERSION = '0.13'


class TestEntry(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Entry()
        obj.id = lexicon.RefId('unique-entry-and-sense-id')
        obj.lexical_unit = lexicon.Multitext()
        obj.sense_items = [lexicon.Sense()]

        xml = f"""
        <entry id="unique-entry-and-sense-id">
            <lexical-unit/>
            <sense/>
        </entry>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestEtymology(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Etymology(
            etym_type=datatypes.Key("Etype"),
            source="Esource"
        )
        obj.gloss_items = [
            base.Gloss(),
            base.Gloss(),
        ]
        obj.form = base.Form()

        xml = f"""
        <etymology type="Etype" source="Esource">
            <gloss/>
            <gloss/>
            <form/>
        </etymology>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestExample(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Example()
        obj.source = base.Key('example source')
        obj.translation_items = [
            lexicon.Translation()
        ]

        xml = f"""
        <example source="example source">
            <translation/>
        </example>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestExtensible(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Etymology()
        obj.type = datatypes.Key("Etype")
        obj.source = "Esource"
        obj.gloss_items = [
            base.Gloss(),
            base.Gloss(),
        ]
        obj.form = base.Form()
        obj.date_created = datatypes.DateTime("0000-00-00")
        obj.date_modified = datatypes.DateTime("0000-00-00")
        obj.field_items = [
            base.Field(),
            base.Field(),
        ]
        obj.trait_items = [
            base.Trait(),
            base.Trait(),
        ]
        obj.annotation_items = [
            base.Annotation(),
            base.Annotation(),
        ]

        xml = f"""
        <etymology dateCreated="{obj.date_created}" dateModified="{obj.date_modified}" type="Etype" source="Esource">
            <field/>
            <field/>
            <trait/>
            <trait/>
            <annotation/>
            <annotation/>
            <gloss/>
            <gloss/>
            <form/>
        </etymology>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestGrammaticalInfo(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.GrammaticalInfo(value=datatypes.Key("g-i-value"))
        obj.trait_items = [
            base.Trait(),
        ]

        xml = f"""
        <grammatical-info value="g-i-value">
            <trait/>
        </grammatical-info>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestLexicon(unittest.TestCase):
    def test_xml(self):
        # config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Lexicon(version='0.13')
        obj.producer = f"LIFTUtils {config.LIB_VERSION}"
        obj.entry_items = [
            lexicon.Entry(),
        ]

        xml = f"""
        <lift version="0.13" producer="LIFTUtils {config.LIB_VERSION}">
            <entry/>
        </lift>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestNote(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Note()
        obj.type = datatypes.Key("note-type")
        obj.form_items = [
            base.Form(
                lang=datatypes.Lang("en"),
                text=base.Text("This is a note")
            )
        ]

        xml = f"""
        <note type="note-type">
            <form lang="en">
                <text>This is a note</text>
            </form>
        </note>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        # compare_xml_trees(obj._to_xml_tree(), xml_tree)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestPhonetic(unittest.TestCase):
    def test_xml(self):
        href1 = 'https://example.com'
        href2 = 'file:///home/user/audio.wav'
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Phonetic()
        obj.media_items = [
            base.URLRef(href1),
            base.URLRef(href2),
        ]
        obj.form_items = [
            base.Span(),
            base.Span(),
        ]

        xml = f"""
        <pronunciation>
            <form/>
            <form/>
            <media href="{href1}"/>
            <media href="{href2}"/>
        </pronunciation>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        # compare_xml_trees(obj._to_xml_tree(), xml_tree)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestRelation(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Relation(
            rel_type=base.Key("rel-type"),
            ref=lexicon.RefId("ref-id")
        )
        obj.usage_items = [
            base.Multitext()
        ]

        xml = f"""
        <relation type="rel-type" ref="ref-id">
            <usage/>
        </relation>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestReversal(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Reversal()
        obj.main = lexicon.Reversal()
        obj.grammatical_info = lexicon.GrammaticalInfo(value="g-i-value")

        xml = f"""
        <reversal>
            <main/>
            <grammatical-info value="g-i-value"/>
        </reversal>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestSense(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Sense()
        obj.grammatical_info = lexicon.GrammaticalInfo()
        obj.definition = lexicon.Multitext()
        obj.subsense_items = [lexicon.Sense()]

        xml = f"""
        <sense>
            <grammatical-info/>
            <definition/>
            <subsense/>
        </sense>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestTranslation(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Translation()
        obj.type = base.Key('literal')
        obj.form_items = [
            base.Form()
        ]

        xml = f"""
        <translation type="literal">
            <form/>
        </translation>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )


class TestVariant(unittest.TestCase):
    def test_xml(self):
        config.LIFT_VERSION = LIFT_VERSION
        obj = lexicon.Variant()
        obj.pronunciation_items = [lexicon.Phonetic()]
        obj.relation_items = [lexicon.Relation()]

        xml = f"""
        <variant>
            <pronunciation/>
            <relation/>
        </variant>
        """
        xml_tree = utils.xmlstring_to_etree(xml)
        self.assertEqual(
            utils.etree_to_xmlstring(obj._to_xml_tree()),
            utils.etree_to_xmlstring(xml_tree)
        )



import unittest
from lxml import etree

from . import DATA_PATH
from .utils import test_attribs
from .utils import test_elems

from lift_utils import config
from lift_utils import lexicon

LIFT_GOOD = str(DATA_PATH / "lexicon_good_v0.13_FW.lift")
LIFT_VERSION = '0.13'


class TestEntry(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//entry')
        self.obj = lexicon.Entry(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestEtymology(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//etymology')
        self.obj = lexicon.Etymology(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestExample(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//example')
        self.obj = lexicon.Example(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')
        test_elems(self, self.obj, optional)


class TestGrammaticalInfo(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//grammatical-info')  # noqa: E501
        self.obj = lexicon.GrammaticalInfo(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestLexicon(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot()
        self.obj = lexicon.Lexicon(path=LIFT_GOOD, xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)

    def test_entry_items(self):
        self.assertTrue(len(self.obj.entry_items) > 0)

    def test_update_header_from_file(self):
        self.assertTrue(len(self.obj.header.ranges.range_items) > 1)


class TestNote(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//note')
        self.obj = lexicon.Note(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        pass  # no elements


class TestPhonetic(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//pronunciation')  # noqa: E501
        self.obj = lexicon.Phonetic(xml_tree=self.xml_tree)

    def test_attribs(self):
        pass  # no attributes

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')
        test_elems(self, self.obj, optional)


class TestRelation(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//relation')
        self.obj = lexicon.Relation(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestReversal(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//reversal')
        self.obj = lexicon.Reversal(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')
        test_elems(self, self.obj, optional)


class TestSense(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//sense')
        self.obj = lexicon.Sense(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestTranslation(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//translation')
        self.obj = lexicon.Translation(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        pass  # no elements


class TestVariant(unittest.TestCase):
    def setUp(self):
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find('.//variant')
        self.obj = lexicon.Variant(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')
        test_elems(self, self.obj, optional)



"""Manipulate lexicon entries and their dependent elements."""

from lxml import etree
from pathlib import Path
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import unquote
from urllib.parse import urlparse

from . import config
from . import utils
from .base import Extensible
from .base import Form
from .base import Gloss
from .base import LIFTUtilsBase
from .base import Multitext
from .base import Span
from .base import Trait
from .base import URLRef
from .datatypes import DateTime
from .datatypes import Key
from .datatypes import RefId
from .datatypes import Prop
from .datatypes import URL
from .header import Header
from .header import Range
from .utils import xmlfile_to_etree
from .utils import etree_to_obj_attributes
from .utils import search_entry


class Note(Multitext, Extensible):
    """For storing descriptive information of many kinds.
    It can include comments, bibliographic information and domain specific
    notes.

    :ivar Optional[Key] type: Gives the type of note by reference to a
        ``range-element`` in the ``note-type`` range.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        Extensible.__init__(self)
        Multitext.__init__(self)
        # properties
        self.props.add_to('attributes', Prop('type', prop_type=Key))
        self.xml_tag = 'note'
        # attributes
        self.type: Optional[Key] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return super().__str__()


class Phonetic(Multitext, Extensible):
    """This represents a single pronunciation in phonetic form.

    :ivar Optional[List[URLRef]] media_items: Stores an audio representation of
        the text.
    :ivar Optional[List[Span]] form_items: `Used by LIFT v0.13 (FieldWorks).`
        Stores the phonetic representation using whichever writing system: IPA,
        Americanist, etc.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        Extensible.__init__(self)
        Multitext.__init__(self)
        # properties
        self.props.add_to(
            'elements',
            Prop('media_items', prop_type=list, item_type=URLRef)
        )
        if config.LIFT_VERSION == '0.13':
            self.props.add_to(
                'elements',
                Prop('form_items', prop_type=list, item_type=Span)
            )
        self.xml_tag = 'pronunciation'
        # elements
        self.media_items: Optional[List[URLRef]] = None
        if config.LIFT_VERSION == '0.13':
            self.form_items: Optional[List[Span]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return super().__str__()


class Etymology(Extensible):
    """For describing lexical relations with a word not in the lexicon.

    :ivar Key type: Gives the etymological relationship between this sense and
        some other word in another language.
    :ivar str source: Gives the language for the source language of the
        etymological relation.
    :ivar Optional[List[Gloss]] gloss_items: Gives glosses of the word that the
        etymological relationship is with.
    :ivar Optional[Form] form: Holds the form of the etymological reference.
    """

    def __init__(
        self,
        etym_type: Key = None,
        source: str = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('type', required=True, prop_type=Key),
            Prop('source', required=True, prop_type=str),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('gloss_items', prop_type=list, item_type=Gloss),
            Prop('form', prop_type=Form),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'etymology'
        # attributes
        self.type = etym_type
        self.source = source
        # elements
        self.gloss_items: Optional[List[Gloss]] = None
        self.form: Optional[Form] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return f"{self.type} ({self.source})"


class GrammaticalInfo(LIFTUtilsBase):
    """A reference to a ``range-element`` in the ``grammatical-info`` range.

    :ivar Key value: The part of speech tag into the ``grammatical-info``
        range.
    :ivar Optional[List[Trait]] trait_items: Allows the grammatical information
        for a given sense to have more information than just the part of speech
        given by the ``value`` attribute.
    """

    def __init__(
        self,
        value: Key = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'attributes',
            Prop('value', required=True, prop_type=Key)
        )
        self.props.add_to(
            'elements',
            Prop('trait_items', prop_type=list, item_type=Trait)
        )
        self.xml_tag = 'grammatical-info'
        # attributes
        self.value = value
        # elements
        self.trait_items: Optional[List[Trait]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        traits = ''
        if self.trait_items:
            traits = f": {'; '.join([str(t) for t in self.trait_items])}"
        return f"{self.value}{traits}"


class Reversal(Multitext):
    """Enables a wider use of a dictionary.

    :ivar Optional[Key] type: Gives the type of the reversal as a
        ``range-element`` in the ``reversal-type`` range.
    :ivar Optional[Reversal] main: This gives the parent reversal in a
        hierarchical set of reversals.
    :ivar Optional[GrammaticalInfo] grammatical_info: This allows a reversal
        relation to specify what the grammatical information is in the reversal
        language.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'attributes',
            Prop('type', prop_type=Key)
        )
        elems = [
            Prop('main', prop_type=Reversal),
            Prop('grammatical_info', prop_type=GrammaticalInfo),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'reversal'
        # attributes
        self.type: Optional[Key] = None
        # elements
        self.main: Optional[Reversal] = None
        self.grammatical_info: Optional[GrammaticalInfo] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Translation(Multitext):
    """A ``Multitext`` with an optional translation ``type`` attribute.

    :ivar Optional[Key] type: Gives the type of the translation.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'attributes',
            Prop('type', prop_type=Key)
        )
        self.xml_tag = 'translation'
        # attributes
        self.type: Optional[Key] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Example(Multitext, Extensible):
    """Gives an example sentence or phrase.
    It is given in the language and glosses of that example in other languages.

    :ivar Optional[Key] source: Reference by which another application may
        refer to this example or is a reference into another database of texts,
        for example.
    :ivar Optional[List[Translation]] translation_items: Gives translations of
        the example into different languages.
    :ivar Optional[List[Note]] note_items: Holds notes on this example.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        Extensible.__init__(self)
        Multitext.__init__(self)
        # properties
        self.props.add_to(
            'attributes',
            Prop('source', prop_type=Key)
        )
        self.props.add_to(
            'elements',
            Prop('translation_items', prop_type=list, item_type=Translation)
        )
        if config.LIFT_VERSION in ['0.15']:
            self.props.add_to(
                'elements',
                Prop('note_items', prop_type=list, item_type=Note)
            )
        self.xml_tag = 'example'
        # attributes
        self.source: Optional[Key] = None
        # elements
        self.translation_items: Optional[List[Translation]] = None
        if config.LIFT_VERSION in ['0.15']:
            self.note_items: Optional[List[Note]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Relation(Extensible):
    """This element is used for lexical relations.

    :ivar Key type: Is the type of the particular lexical relation.
    :ivar RefId ref: This is the other end of the relation, either a ``sense``
        or an ``entry``.
    :ivar Optional[int] order: Gives the relative ordering of relations of a
        given type when a multiple relation is being described.
    :ivar Optional[Multitext] usage: Gives information on usage in one or more
        languages or writing systems.
    """

    def __init__(
        self,
        rel_type: Key = None,
        ref: RefId = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('type', required=True, prop_type=Key),
            Prop('ref', required=True, prop_type=RefId),
            Prop('order', prop_type=int),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        self.props.add_to(
            'elements',
            Prop('usage_items', prop_type=list, item_type=Multitext)
        )
        self.xml_tag = 'relation'
        # attributes
        self.type = rel_type
        self.ref = ref
        self.order: Optional[int] = None
        # elements
        self.usage_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return f"{self.type}: {self.ref}"


class Variant(Multitext, Extensible):
    """``variant`` elements are used for all sorts of variation.

    :ivar Optional[RefId] ref: Gives the variation as a reference to another
        ``entry`` or ``sense`` rather than specifying the ``form`` (that is,
        the ``Multitext`` value of the variant).
    :ivar Optional[List[Phonetic]] pronunciation_items: Holds the phonetic
        variant whether it is that this is a variation in phonetics only or
        that the phonetic variation arises because of an orthographic or
        phonemic variation.
    :ivar Optional[List[Relation]] relation_items: Some variants have a lexical
        relationship with other senses or entries in the lexicon.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        Extensible.__init__(self)
        Multitext.__init__(self)
        # properties
        self.props.add_to(
            'attributes',
            Prop('ref', prop_type=RefId)
        )
        elems = [
            Prop('pronunciation_items', prop_type=list, item_type=Phonetic),
            Prop('relation_items', prop_type=list, item_type=Relation),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'variant'
        # attributes
        self.ref: Optional[RefId] = None
        # elements
        self.pronunciation_items: Optional[List[Phonetic]] = None
        self.relation_items: Optional[List[Relation]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return self.ref if self.ref else 'variant'


class Sense(Extensible):
    """An ``entry`` is made up of a number of ``sense`` elements.

    :ivar Optional[RefId] id: This gives an identifier for this ``Sense`` so
        that things can refer to it.
    :ivar Optional[int] order: A number that is used to give the relative
        order of senses within an entry.
    :ivar Optional[GrammaticalInfo] grammatical_info: Grammatical information.
    :ivar Optional[List[Union[Gloss, Form]]] gloss_items: Each ``gloss`` is a
        single string in a single language and writing system. ``Form`` is used
        by LIFT v0.13 (FieldWorks), while ``Gloss`` is used in later versions.
    :ivar Optional[Multitext] definition: Gives the definition in multiple
        languages or writing systems.
    :ivar Optional[List[Relation]] relation_items: While a lexical relation
        isn't strictly owned by a sense it is a good place to hold it.
    :ivar Optional[List[Note]] note_items: There are lots of different types of
        notes.
    :ivar Optional[List[Example]] example_items: Examples may be used for
        different target audiences.
    :ivar Optional[List[Reversal]] reversal_items: There may be different
        reversal indexes.
    :ivar Optional[List[URLRef]] illustration_items: The picture doesn't have
        to be static.
    :ivar Optional[List[Sense]] subsense_items: Senses can form a hierarchy.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('id', prop_type=RefId),
            Prop('order', prop_type=int),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('grammatical_info', prop_type=GrammaticalInfo),
            Prop('definition', prop_type=Multitext),
            Prop('relation_items', prop_type=list, item_type=Relation),
            Prop('note_items', prop_type=list, item_type=Note),
            Prop('example_items', prop_type=list, item_type=Example),
            Prop('reversal_items', prop_type=list, item_type=Reversal),
            Prop('illustration_items', prop_type=list, item_type=URLRef),
            Prop('subsense_items', prop_type=list, item_type=Sense),
        ]
        if config.LIFT_VERSION == '0.13':
            elems.append(Prop('gloss_items', prop_type=list, item_type=Form))
        else:
            elems.append(Prop('gloss_items', prop_type=list, item_type=Gloss))
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'sense'
        # attributes
        self.id: Optional[RefId] = None
        self.order: Optional[int] = None
        # elements
        self.grammatical_info: Optional[GrammaticalInfo] = None
        if config.LIFT_VERSION == '0.13':
            self.gloss_items: Optional[List[Form]] = None
        else:
            self.gloss_items: Optional[List[Gloss]] = None
        self.definition: Optional[Multitext] = None
        self.relation_items: Optional[List[Relation]] = None
        self.note_items: Optional[List[Note]] = None
        self.example_items: Optional[List[Example]] = None
        self.reversal_items: Optional[List[Reversal]] = None
        self.illustration_items: Optional[List[URLRef]] = None
        self.subsense_items: Optional[List] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return self._summary_line()

    def _summary_line(self, lang='en'):
        """Return a one-line summary of the entry's data for a given language.
        Defaults to English.
        """
        gl = utils.ellipsize(str(self.get_gloss(lang=lang)), 20)
        gi = utils.ellipsize(self.get_grammatical_info(), 10)
        return f"{gl:20}\t{gi:10}\t{self.id}"

    def get_id(self) -> RefId:
        """Return the object's unique identifier"""
        return self.id

    def get_gloss(self, lang='en') -> str:
        """Get gloss details for a given language.
        Defaults to English.
        """
        gloss = ''
        if self.gloss_items:
            # Choose 1st gloss if preferred language not found.
            gloss = str(self.gloss_items[0])
            for g in self.gloss_items:
                if g.lang == lang:
                    # Choose preferred language gloss.
                    gloss = str(g)
        return gloss

    def get_grammatical_info(self) -> GrammaticalInfo:
        """Get basic grammatical info.
        """
        grammatical_info = ''
        if self.grammatical_info:
            grammatical_info = str(self.grammatical_info)
        return grammatical_info

    def show(self):
        """Print an overview of the ``sense`` in the terminal window."""
        print(self)


class Entry(Extensible):
    """This is the core of a lexicon. A lexicon is made up of a set of entries.

    :ivar Optional[RefId] id: This gives a unique identifier to this ``entry``.
    :ivar Optional[str] guid: `Deprecated.` This gives a unique identifier to
        this entry in the form of a “universally unique identifier” (RFC 4122).
    :ivar Optional[int] order: This is the homograph number.
    :ivar Optional[DateTime] date_deleted: If this attribute exists then it
        indicates that the particular ``entry`` has been deleted.
    :ivar Optional[Multitext] lexical_unit: The lexical form is the primary
        lexical form as is found as the primary lexical form in the source data
        models for this standard.
    :ivar Optional[Multitext] citation: This is the form that is to be printed
        in the dictionary.
    :ivar Optional[List[Phonetic]] pronunciation_items: There can be multiple
        phonetic forms of an entry.
    :ivar Optional[List[Variant]] variant_items: Any constrained variants or
        free orthographic variants.
    :ivar Optional[List[Sense]] sense_items: This is where the definition goes.
    :ivar Optional[List[Note]] note_items: The more notes you keep the better.
    :ivar Optional[List[Relation]] relation_items: Gives a lexical relationship
        between this entry and another ``entry`` or ``sense``.
    :ivar Optional[List[Etymology]] etymology_items: Differs from a lexical
        relation in that it has no referent in the lexicon.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('id', prop_type=RefId),
            Prop('guid', prop_type=str),
            Prop('order', prop_type=str),
            Prop('date_deleted', prop_type=DateTime),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('lexical_unit', prop_type=Multitext),
            Prop('citation', prop_type=Multitext),
            Prop('pronunciation_items', prop_type=list, item_type=Phonetic),
            Prop('variant_items', prop_type=list, item_type=Variant),
            Prop('sense_items', prop_type=list, item_type=Sense),
            Prop('note_items', prop_type=list, item_type=Note),
            Prop('relation_items', prop_type=list, item_type=Relation),
            Prop('etymology_items', prop_type=list, item_type=Etymology),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'entry'
        # attributes
        self.id: Optional[RefId] = None
        self.guid: Optional[str] = None  # deprecated
        self.order: Optional[int] = None
        self.date_deleted: Optional[DateTime] = None
        # elements
        self.lexical_unit: Optional[Multitext] = None
        self.citation: Optional[Multitext] = None
        self.pronunciation_items: Optional[List[Phonetic]] = None
        self.variant_items: Optional[List[Variant]] = None
        self.sense_items: Optional[List[Sense]] = None
        self.note_items: Optional[List[Note]] = None
        self.relation_items: Optional[List[Relation]] = None
        self.etymology_items: Optional[List[Etymology]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return self._summary_line()

    def _summary_line(self, lang='en'):
        """Return a one-line summary of the entry's data for a given language.
        Defaults to English.
        """
        lu = utils.ellipsize(str(self.lexical_unit), 20)
        gl = self.get_gloss(lang=lang)
        return f"{lu:20}\t{gl:30}\t{self.id}"

    def get_id(self) -> RefId:
        """Return the object's unique identifier"""
        return self.id

    def get_grammatical_info(self, sense_idx=0) -> GrammaticalInfo:
        """Get basic grammatical info for a given sense index [default=0].
        """
        grammatical_info = ''
        if self.sense_items:
            grammatical_info = self.sense_items[sense_idx].get_grammatical_info()  # noqa: E501
        return grammatical_info

    def get_gloss(self, sense_idx=0, lang='en') -> str:
        """Get gloss details for a given sense index and language.
        Defaults to index 0 and English.
        """
        gloss = ''
        if self.sense_items and self.sense_items[sense_idx].gloss_items:
            # Choose 1st gloss if preferred language not found.
            gloss = str(self.sense_items[sense_idx].gloss_items[0])
            for g in self.sense_items[sense_idx].gloss_items:
                if g.lang == lang:
                    # Choose preferred language gloss.
                    gloss = str(g)
        return gloss

    def show(self):
        """Print an overview of the ``entry`` in the terminal window."""
        # Add header.
        text = ['\nEntry\n============']
        # Add overview line.
        if self.lexical_unit:
            text.append(f"{self.lexical_unit}; {self.get_grammatical_info()}; {self.get_gloss()}")  # noqa: E501
        # Add traits.
        if self.trait_items:
            text.append('; '.join([str(t) for t in self.trait_items]))
        # Add senses.
        if self.sense_items:
            text.append('\n'.join([f"sense: {str(s)}" for s in self.sense_items]))  # noqa: E501
        # Add variants.
        if self.variant_items:
            text.append('; '.join(str(v) for v in self.variant_items))
        # Add notes.
        if self.note_items:
            text.append('; '.join(str(n) for n in self.note_items))
        # Add etymologies.
        if self.etymology_items:
            text.append('; '.join(str(e) for e in self.etymology_items))
        print('\n'.join(text))


class Lexicon(LIFTUtilsBase):
    """This is the main class of the lexicon.
    It contains the header and all the entries in the database.

    :ivar str version: Specifies the lift language version number.
    :ivar Optional[str] producer: Identifies the particular producer of this
        lift file.
    :ivar Optional[Header] header: Contains the header information for the
        database.
    :ivar Optional[List[Entry]] entry_items: Each of the entries in the
        lexicon.
    :ivar Optional[Path] path: File path to a LIFT file to import.
    """

    _producer = f"LIFT-Utils {config.LIB_VERSION}"

    def __init__(
        self,
        path: Optional[Union[Path, str]] = None,
        version: str = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('version', required=True, prop_type=str),
            Prop('producer', prop_type=str),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('header', prop_type=Header),
            Prop('entry_items', prop_type=list, item_type=Entry),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'lift'
        self.lift_xml_tree = None
        self.ranges_xml_tree = None
        # attributes
        self.version = version
        self.producer: Optional[str] = None
        # elements
        self.header: Optional[Header] = None
        self.entry_items: Optional[List[Entry]] = None
        if path:
            self.path = Path(path)
            if self.path.suffix == '.lift':
                self._from_lift(self.path)
            else:
                print(f"Error: Not a valid LIFT file: {self.path}")
                exit(1)
        elif xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        s = f"LIFT lexicon v{self.version}"
        if self.producer:
            s += f"; produced by {self.producer}"
        return s

    def find(
        self,
        text: str,
        field: str = 'gloss'
    ) -> Union[Entry, Sense, None]:
        """Return the first matching ``Entry`` or ``Sense`` item.
        The field searched can be "lexical-unit", "variant", "gloss", or
        "definition", as well as any fields defined in the LIFT file's header.
        """
        return self._find(text, field=field)

    def find_all(
        self,
        text: str,
        field: str = 'gloss'
    ) -> List[Union[Entry, Sense]]:
        """Return all matching ``Entry`` or ``Sense`` items.
        The field searched can be "lexical-unit", "variant", "gloss",
        "definition", or "grammatical-info", as well as any fields defined in
        the LIFT file's header.
        """
        return self._find(text, field=field, get_all=True)

    def _find(self, text, field='gloss', get_all=False):
        items = []

        target_groups = ['entries', 'senses']
        entry_only_fields = ['lexical-unit', 'variant']
        sense_only_fields = ['gloss', 'definition', 'grammatical-info']
        header_fields = [f.tag for f in self.header.fields.field_items]
        if field in entry_only_fields:
            target_groups.remove('senses')
        elif field in sense_only_fields:
            target_groups.remove('entries')

        for entry in self.entry_items:
            result = search_entry(
                entry,
                text,
                field,
                target_groups,
                header_fields,
                get_all
            )
            if result is not None:
                if not get_all:
                    return result
                else:
                    items.extend(result)
        if get_all:
            return items

    def get_item_by_id(self, refid: str) -> Union[Entry, Sense, None]:
        """Return an entry or sense by its ``RefId``."""
        if not self.entry_items:
            return
        for entry in self.entry_items:
            if entry.id == refid:
                return entry
            if entry.sense_items:
                for sense in entry.sense_items:
                    if sense.id == refid:
                        return sense
                    if sense.subsense_items:
                        for subsense in sense.subsense_items:
                            if subsense.id == refid:
                                return subsense

    def show(self):
        """Print an overview of the ``Lexicon`` in the terminal window."""
        text = "No entries."
        if self.entry_items:
            summary_lines = [e._summary_line('en') for e in self.entry_items]  # noqa: E501
            slist = utils.unicode_sort(summary_lines)
            nl = '\n'
            text = nl.join(slist)
        print(text)

    def _from_lift(self, infile):
        infile = Path(infile)
        if not infile.is_file():
            print(f"Error: Invalid file path: {infile}")
            exit(1)
        xml_tree = xmlfile_to_etree(infile)
        self._update_from_xml(xml_tree)

    def to_lift(self, file_path: str):
        """Save the ``Lexicon`` as a LIFT file.
        The LIFT-RANGES file will be automatically created in the same folder
        as the LIFT file.
        """
        outfile = Path(file_path).with_suffix('.lift')  # ensure suffix
        ranges_file = outfile.with_suffix('.lift-ranges')
        self.producer = Lexicon._producer

        # Write LIFT file.
        lift_tree = self._to_xml_tree()
        for _range in lift_tree.find('.//ranges').getchildren():
            for _range_element in _range:
                _range.remove(_range_element)
            for attrib in _range.attrib.keys():
                if attrib not in ['id', 'href']:
                    del _range.attrib[attrib]
        outfile.write_text(self._to_xml(lift_tree))

        # Write LIFT-RANGES file.
        lift_ranges = self.header.ranges._to_xml_tree()
        lift_ranges.tag = 'lift-ranges'
        for _range in list(lift_ranges):
            del _range.attrib['href']
        ranges_file.write_text(self._to_xml(lift_ranges))

    def _update_from_xml(self, xml_tree):
        # Set initial xml_tree.
        self.xml_tree = xml_tree
        self.version = xml_tree.attrib.get('version')
        # Allow global access to version number.
        config.LIFT_VERSION = self.version
        # Update object attributes.
        etree_to_obj_attributes(xml_tree, self)
        # Update header range data from external file(s).
        ext_hrefs = set()
        for r in self.header.ranges.range_items:
            if r.href:
                ext_hrefs.add(r.href)
        for p in ext_hrefs:
            self._update_header_from_href(p)
        # Update ranges xml_trees.
        self.header.ranges._to_xml_tree()

    def _update_header_from_href(self, href: URL):
        filepath = unquote(urlparse(href).path)
        try:
            xml_tree = xmlfile_to_etree(filepath)
        except OSError:
            # Probably absolute URI from a different device.
            # Try same file name, but in same dir as current LIFT file.
            relpath = self.path.parent / Path(filepath).name
            xml_tree = xmlfile_to_etree(relpath)

        # Update model data.
        for _range in xml_tree.getchildren():
            for i, r in enumerate(self.header.ranges.range_items[:]):
                if _range.attrib.get('id') == r.id:
                    href = r.href
                    self.header.ranges.range_items[i] = Range(xml_tree=_range)
                    self.header.ranges.range_items[i].href = href  # add href
                    break



import unittest
from lxml import etree

from . import DATA_PATH
from .utils import test_attribs
from .utils import test_elems

from lift_utils import base
from lift_utils import config

ENTRY_LIFT_GOOD = str(DATA_PATH / "entry_good_v0.15.lift")


class TestAnnotation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//annotation')  # noqa: E501
        self.obj = base.Annotation(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        pass  # no elements


class TestExtensible(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.obj = base.Extensible(etree.parse(ENTRY_LIFT_GOOD).getroot())

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestField(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//field')
        self.obj = base.Field(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')
        test_elems(self, self.obj, optional)


class TestForm(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//form')
        self.obj = base.Form(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestGloss(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//gloss')
        self.obj = base.Gloss(xml_tree=self.xml_tree)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestMultitext(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//annotation')  # noqa: E501
        self.obj = base.Multitext(xml_tree=self.xml_tree)

    def test_attribs(self):
        pass  # no attribs

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')
        test_elems(self, self.obj, optional)


class TestSpan(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//span')
        self.obj = base.Span(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')  # hard to test optional nested element
        test_elems(self, self.obj, optional)


class TestText(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//form')
        self.obj = None
        for c in self.xml_tree.getchildren():
            if c.tag == 'text':
                self.obj = base.Text(c)
                break
        if not self.obj:
            raise AssertionError

    def test_attribs(self):
        pass  # no attribs

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('span_items')  # already tested in TestSpan
        test_elems(self, self.obj, optional)


class TestTrait(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('trait')
        self.obj = base.Trait(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestURLRef(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = config.LIFT_VERSION_LATEST
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find('.//urlref')  # noqa: E501
        self.obj = base.URLRef(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)



import unittest
from lxml import etree

from . import DATA_PATH
from .utils import test_attribs
from .utils import test_elems

from lift_utils import config
from lift_utils import header

HEADER_LIFT_GOOD = str(DATA_PATH / "header_good_v0.15.lift")
LIFT_VERSION = '0.15'


class TestFieldDefinition(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find('.//field')  # noqa: E501
        self.obj = header.FieldDefinition(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestFields(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find('.//fields')  # noqa: E501
        self.obj = header.Fields(xml_tree=self.xml_tree)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)

    def test_items(self):
        self.assertTrue(len(self.obj.field_items) > 1)


class TestHeader(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find('.//header')  # noqa: E501
        self.obj = header.Header(xml_tree=self.xml_tree)

    def test_attribs(self):
        pass  # no attributes to test

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)


class TestRanges(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find('.//ranges')  # noqa: E501
        self.obj = header.Ranges(xml_tree=self.xml_tree)

    def test_attribs(self):
        pass  # no attributes

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        test_elems(self, self.obj, optional)

    def test_items(self):
        self.assertTrue(len(self.obj.range_items) > 1)


class TestRange(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find('.//range')  # noqa: E501
        self.obj = header.Range(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('annotation_items')
        optional.remove('field_items')
        optional.remove('trait_items')
        test_elems(self, self.obj, optional)


class TestRangeElement(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find('.//range-element')  # noqa: E501
        self.obj = header.RangeElement(xml_tree=self.xml_tree)

    def test_attribs(self):
        required = [p.name for p in self.obj.props.attributes if p.required]
        test_attribs(self, self.obj, required)
        optional = [p.name for p in self.obj.props.attributes if not p.required]  # noqa: E501
        test_attribs(self, self.obj, optional)

    def test_elems(self):
        required = [p.name for p in self.obj.props.elements if p.required]
        test_elems(self, self.obj, required)
        optional = [p.name for p in self.obj.props.elements if not p.required]
        optional.remove('annotation_items')
        optional.remove('field_items')
        optional.remove('trait_items')
        test_elems(self, self.obj, optional)



"""Define runtime global variables."""

LIB_VERSION = '0.2'
LIFT_VERSION_FIELDWORKS = '0.13'  # used by FLEx 9.1
LIFT_VERSION_LATEST = '0.15'
LIFT_VERSION_DEFAULT = LIFT_VERSION_FIELDWORKS
LIFT_VERSION = None

XML_NAMES = {
    # object attribute keys (L) and corresponding LIFT XML tags (R)
    'abbrev_items': 'abbrev',
    'annotation_items': 'annotation',
    'class_': 'class',
    'date_created': 'dateCreated',
    'date_deleted': 'dateDeleted',
    'date_modified': 'dateModified',
    'description_items': 'description',
    'entry_items': 'entry',
    'etymology_items': 'etymology',
    'example_items': 'example',
    'field_items': 'field',
    'form_items': 'form',
    'gloss_items': 'gloss',
    'grammatical_info': 'grammatical-info',
    'illustration_items': 'illustration',
    'label_items': 'label',
    'lexical_unit': 'lexical-unit',
    'media_items': 'media',
    'note_items': 'note',
    'option_range': 'option-range',
    'pcdata': 'text',
    'pronunciation_items': 'pronunciation',
    'prop_type': 'type',
    'range_element_items': 'range-element',
    'range_items': 'range',
    'relation_items': 'relation',
    'reversal_items': 'reversal',
    'sense_items': 'sense',
    'span_items': 'span',
    'subsense_items': 'subsense',
    'trait_items': 'trait',
    'translation_items': 'translation',
    'usage_items': 'usage',
    'variant_items': 'variant',
    'writing_system': 'writing-system',
}



"""Manipulate the header section."""

from lxml import etree
from typing import List
from typing import Optional

from . import config
from .base import Extensible
from .base import LIFTUtilsBase
from .base import Multitext
from .datatypes import Key
from .datatypes import Prop
from .datatypes import URL


class FieldDefn(Multitext):
    """Gives information about a particular field type.
    It may be used by an application to add information not part of the LIFT
    standard.

    .. note:: Used by LIFT v0.13 (FieldWorks) instead of ``FieldDefinition``.
    """

    def __init__(
        self,
        tag: Key = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'attributes',
            Prop('tag', required=True, prop_type=Key)
        )
        self.xml_tag = 'field'
        # attributes
        self.tag = tag

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        forms = 'forms' if len(self.form_items) > 1 else 'form'
        return f"{self.tag} ({len(self.form_items)} {forms})"


class FieldDefinition(LIFTUtilsBase):
    """Gives information about a particular field type.
    It may be used by an application to add information not part of the LIFT
    standard.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('name', required=True, prop_type=Key),
            Prop('class_', prop_type=str),
            Prop('type', prop_type=str),
            Prop('option_range', prop_type=Key),
            Prop('writing_system', prop_type=str),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('label', prop_type=Multitext),
            Prop('description', prop_type=Multitext),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'field'
        # attributes
        self.name: Key = None
        self.class_: Optional[str] = None
        self.type: Optional[str] = None
        self.option_range: Optional[Key] = None
        self.writing_system: Optional[str] = None
        # elements
        self.label: Optional[Multitext] = None
        self.description: Optional[Multitext] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return self.name


class FieldDefns(LIFTUtilsBase):
    """This is a simple list of ``field-defn`` elements.

    .. note:: Used by LIFT v0.13 (FieldWorks) instead of ``Fields``.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'elements',
            Prop('field_items', prop_type=list, item_type=FieldDefn)
        )
        self.xml_tag = 'fields'
        # elements
        self.field_items: Optional[List[FieldDefn]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Fields(LIFTUtilsBase):
    """This is a simple list of ``field-definition`` elements.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'elements',
            Prop('field_items', prop_type=list, item_type=FieldDefinition)  # noqa: E501
        )
        # elements
        self.field_items: Optional[List[FieldDefinition]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class RangeElement(Extensible, LIFTUtilsBase):
    """The description of a particular range element found in a ``range``.

    .. note:: Does not inherit from Extensible in LIFT v0.13 (FieldWorks).
    """

    def __init__(
        self,
        elem_id: Key = None,
        xml_tree: Optional[etree._Element] = None
    ):
        if config.LIFT_VERSION == '0.13':
            LIFTUtilsBase.__init__(self)
        else:
            Extensible.__init__(self)
        # properties
        attribs = [
            Prop('id', required=True, prop_type=Key),
            Prop('parent', prop_type=Key),
            Prop('guid', prop_type=str),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('description_items', prop_type=list, item_type=Multitext),
            Prop('label_items', prop_type=list, item_type=Multitext),
            Prop('abbrev_items', prop_type=list, item_type=Multitext),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'range-element'
        # attributes
        self.id = elem_id
        self.parent: Key = None
        self.guid: str = None
        # elements
        self.description_items: Optional[List[Multitext]] = None
        self.label_items: Optional[List[Multitext]] = None
        self.abbrev_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Range(Extensible, LIFTUtilsBase):
    """A set of ``range-elements``.
    It is used to identify both the group of ``range-elements`` but also to
    some extent their type.

    .. note:: Does not inherit from Extensible in LIFT v0.13 (FieldWorks).
    """

    def __init__(
        self,
        range_id: Key = None,
        xml_tree: Optional[etree._Element] = None
    ):
        if config.LIFT_VERSION == '0.13':
            LIFTUtilsBase.__init__(self)
        else:
            Extensible.__init__(self)
        # properties
        attribs = [
            Prop('id', required=True, prop_type=Key),
            Prop('guid', prop_type=str),
            Prop('href', prop_type=URL),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop(
                'range_element_items',
                required=True,
                prop_type=list,
                item_type=RangeElement
            ),
            Prop('description', prop_type=Multitext),
            Prop('label_items', prop_type=list, item_type=Multitext),
            Prop('abbrev_items', prop_type=list, item_type=Multitext),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'range'
        # attributes
        self.id = range_id
        self.guid: Optional[str] = None
        self.href: Optional[URL] = None
        # elements
        self.description: Optional[Multitext] = None
        self.range_element_items: List[RangeElement] = None
        self.label_items: List[Multitext] = None
        self.abbrev_items: List[Multitext] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Ranges(LIFTUtilsBase):
    """The root element in a Lift Ranges file.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'elements',
            Prop('range_items', required=True, prop_type=list, item_type=Range),  # noqa: E501
        )
        self.xml_tag = 'ranges'
        # elements
        self.range_items: List[Range] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Header(LIFTUtilsBase):
    """The header information for a LIFT file.
    It includes range information and added field definitions.

    :ivar Optional[Multitext] description: Contains a multilingual description
        of the lexicon for information purposes only.
    :ivar Optional[Ranges] ranges: Contains all the ``range`` information.
    :ivar Optional[FieldDefns] field_items: `Used by LIFT v0.13 (FieldWorks).`
        Contains definitions for all the ``field`` types used in the document.
    :ivar Optional[Fields] fields: Contains definitions for all the ``field``
        types used in the document.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        elems = [
            Prop('description', prop_type=Multitext),
            Prop('ranges', prop_type=Ranges),
        ]
        if config.LIFT_VERSION == '0.13':
            elems.append(Prop('fields', prop_type=FieldDefns))
        else:
            elems.append(Prop('fields', prop_type=Fields))
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'header'
        # elements
        self.description: Optional[Multitext] = None
        self.ranges: Optional[Ranges] = None
        if config.LIFT_VERSION == '0.13':
            self.fields: Optional[FieldDefns] = None
        else:
            self.fields: Optional[Fields] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        s = 'Header'
        if self.description:
            s = str(self.description)
        if self.ranges:
            s += f": {len(self.ranges)} ranges"
        if self.fields:
            s += f", {len(self.fields)} fields"
        return s



"""Various utility functions."""

import unidecode
from lxml import etree

from . import config


def ellipsize(string, length):
    if len(string) > length:
        i = length - 1
        string = f"{string[:i]}…"
    return string


def etree_to_obj_attributes(xml_tree, obj):
    obj.xml_tree = xml_tree

    if obj.props.attributes:
        for p in obj.props.attributes:
            key = config.XML_NAMES.get(p.name, p.name)
            if key in obj.xml_tree.attrib.keys():
                obj.__dict__[p.name] = p.prop_type(obj.xml_tree.attrib.get(key))  # noqa: E501

    if obj.props.elements:
        for p in obj.props.elements:
            tag = config.XML_NAMES.get(p.name, p.name)
            if obj.xml_tree.text and p.name == 'pcdata':
                obj.__dict__[p.name] = p.prop_type(obj.xml_tree.text)
                continue
            elif obj.xml_tree.tail and p.name == 'tail':
                obj.__dict__[p.name] = p.prop_type(obj.xml_tree.tail)
                continue
            for c in obj.xml_tree.getchildren():
                if c.tag == tag:
                    if hasattr(p.prop_type, 'append'):  # list-like obj/elem
                        c_obj = p.item_type(xml_tree=c)
                        if obj.__dict__.get(p.name) is None:
                            # Instantiate list-like object.
                            obj.__dict__[p.name] = p.prop_type()
                        obj.__dict__[p.name].append(c_obj)
                    else:  # single element
                        obj.__dict__[p.name] = p.prop_type(xml_tree=c)


def obj_attributes_to_etree(obj, root_tag):
    xml_tree = etree.Element(root_tag)

    if obj.props.attributes:
        for attrib in obj.props.attributes:
            attr = attrib.name
            val = obj.__dict__.get(attr)
            name = config.XML_NAMES.get(attr, attr)
            if val is not None:
                xml_tree.set(name, str(val))

    if obj.props.elements:
        for elem in obj.props.elements:
            attr = elem.name
            name = config.XML_NAMES.get(attr, attr)
            val = obj.__dict__.get(attr)
            if hasattr(val, 'append'):  # list-like element
                for o in obj.__dict__.get(attr):
                    xml_tree.append(obj_attributes_to_etree(o, name))
            elif val and attr == 'pcdata':
                xml_tree.text = val
            elif val and attr == 'tail':
                xml_tree.tail = val
            elif val:  # single element
                xml_tree.append(
                    obj_attributes_to_etree(obj.__dict__.get(attr), name)
                )
    return xml_tree


def get_xml_parser():
    return etree.XMLParser(remove_blank_text=True)


def unicode_sort(in_list):
    def fmt(string):
        return unidecode.unidecode(string.lower())
    return sorted(in_list, key=fmt)


def xmlfile_to_etree(filepath):
    return etree.parse(str(filepath), get_xml_parser()).getroot()


def xmlstring_to_etree(xmlstring):
    return etree.fromstring(xmlstring, get_xml_parser())


def etree_to_xmlstring(xml_tree):
    return etree.tostring(
        xml_tree,
        encoding='UTF-8',
        pretty_print=True,
        xml_declaration=True
    ).decode().rstrip()


def search_entry(
    entry,
    text,
    field,
    target_groups,
    header_fields,
    get_all
):
    items = []
    if 'senses' in target_groups:
        if not entry.sense_items:
            if not get_all:
                return
            else:
                return items
        for sense in entry.sense_items:
            result = search_sense(
                sense,
                text,
                field,
                target_groups,
                header_fields,
                get_all
            )
            if result is not None:
                if not get_all:
                    return result
                else:
                    items.extend(result)

            if not sense.subsense_items:
                continue
            for subsense in sense.subsense_items:
                result = search_sense(
                    subsense,
                    text,
                    field,
                    target_groups,
                    header_fields,
                    get_all
                )
                if result is not None:
                    if not get_all:
                        return result
                    else:
                        items.extend(result)

    if 'entries' in target_groups:
        if field == 'lexical-unit':
            if not entry.lexical_unit or not entry.lexical_unit.form_items:  # noqa: E501
                if not get_all:
                    return
                else:
                    return items
            for form in entry.lexical_unit.form_items:
                if text in str(form.text):
                    if not get_all:
                        return entry
                    else:
                        items.append(entry)
        elif field == 'variant':
            if not entry.variant_items:
                if not get_all:
                    return
                else:
                    return items
            for variant in entry.variant_items:
                if not variant.form_items:
                    continue
                for form in variant.form_items:
                    if text in str(form.text):
                        if not get_all:
                            return entry
                        else:
                            items.append(entry)
        elif field in header_fields:
            if not entry.field_items:
                if not get_all:
                    return
                else:
                    return items
            for field_item in entry.field_items:
                if field_item.type == field and field_item.form_items:  # noqa: E501
                    for form in field_item.form_items:
                        if text in str(form.text):
                            if not get_all:
                                return entry
                            else:
                                items.append(entry)
    if get_all:
        return items


def search_sense(
    sense,
    text,
    field,
    target_groups,
    header_fields,
    get_all
):
    items = []
    if field == 'gloss':
        if not sense.gloss_items:
            if not get_all:
                return
            else:
                return items
        for gloss in sense.gloss_items:
            if text in str(gloss.text):
                if not get_all:
                    return sense
                else:
                    items.append(sense)
    elif field == 'definition':
        if not sense.definition or not sense.definition.form_items:  # noqa: E501
            if not get_all:
                return
            else:
                return items
        for form in sense.definition.form_items:
            if text in str(form.text):
                if not get_all:
                    return sense
                else:
                    items.append(sense)
    elif field == 'grammatical-info':
        if not sense.grammatical_info:
            if not get_all:
                return
            else:
                return items
        if text in str(sense.grammatical_info):
            if not get_all:
                return sense
            else:
                items.append(sense)
    elif field in header_fields:
        if not sense.field_items:
            if not get_all:
                return
            else:
                return items
        for field_item in sense.field_items:
            if field_item.type == field and field_item.form_items:  # noqa: E501
                for form in field_item.form_items:
                    if text in str(form.text):
                        if not get_all:
                            return sense
                        else:
                            items.append(sense)
    if get_all:
        return items



import unittest
from lxml import etree

from . import DATA_PATH

from lift_utils import lexicon
from lift_utils.utils import xmlfile_to_etree

LIFT_GOOD = DATA_PATH / "sango" / "sango.lift"
LEXICON = lexicon.Lexicon(LIFT_GOOD)


class TestParse(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON

    def test_lexicon_header_field_items(self):
        self.assertTrue(len(self.lexicon.header.fields.field_items) == 8)

    def test_lexicon_header_ranges(self):
        self.assertTrue(len(self.lexicon.header.ranges.range_items) == 20)

    def test_lexicon_entry_items(self):
        self.assertTrue(len(self.lexicon.entry_items) == 3507)


class TestReadWrite(unittest.TestCase):
    # NOTE: I haven't yet found a reliable way to compare XML file data. Using
    # "canonicalize" and "tostring" both result in XML strings whose child
    # element orders depend on the tree order rather than some kind of sorted
    # order useful for comparison. And since the element order of the input
    # file is unpredictable, there's no way to guarantee that the output file
    # will produce the same order, even if contents are identical.
    def test_read_write_file(self):
        infile = LIFT_GOOD
        outfile = DATA_PATH / "sango_out.lift"
        LEXICON.to_lift(outfile)
        xml_in = etree.tostring(
            xmlfile_to_etree(infile),
            encoding='UTF-8',
            pretty_print=True,
            xml_declaration=True
        ).decode().rstrip()
        xml_out = etree.tostring(
            xmlfile_to_etree(outfile),
            encoding='UTF-8',
            pretty_print=True,
            xml_declaration=True
        ).decode().rstrip()

        ranges_xml_in = etree.tostring(
            xmlfile_to_etree(infile.with_suffix('.lift-ranges')),
            encoding='UTF-8',
            pretty_print=True,
            xml_declaration=True
        ).decode().rstrip()
        ranges_xml_out = etree.tostring(
            xmlfile_to_etree(outfile.with_suffix('.lift-ranges')),
            encoding='UTF-8',
            pretty_print=True,
            xml_declaration=True
        ).decode().rstrip()
        # xml_in = etree.canonicalize(
        #     xmlfile_to_etree(infile),
        #     strip_text=True
        # )
        # xml_out = etree.canonicalize(
        #     xmlfile_to_etree(outfile),
        #     strip_text=True
        # )
        # self.maxDiff = None
        # NOTE: In lieu of being able to compare xml content, just using a file
        # size comparison for the time being. This works okay as long as there
        # aren't too many comments in the infile, since comments aren't kept.
        # This tests that the outfile's size is within 7% of the infile's.
        delta = 0.05
        # print(len(xml_out)/len(xml_in))
        # print(len(ranges_xml_out)/len(ranges_xml_in))
        self.assertAlmostEqual(
            1,
            len(xml_out)/len(xml_in),
            delta=delta
        )
        self.assertAlmostEqual(
            1,
            len(ranges_xml_out)/len(ranges_xml_in),
            delta=delta
        )


class TestFind(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON

    def test_find_all_definition(self):
        self.assertEqual(
            len(self.lexicon.find_all(
                "husband",
                field="definition"
            )), 2)

    def test_find_all_gloss(self):
        self.assertEqual(len(self.lexicon.find_all("head")), 20)

    def test_find_all_grammatical_info(self):
        self.assertEqual(
            len(self.lexicon.find_all("Nom", field="grammatical-info")),
            2381
        )

    def test_find_all_header_field(self):
        self.assertEqual(len(self.lexicon.find_all("1021", field="CAWL")), 1)

    def test_find_definition(self):
        self.assertEqual(
            str(self.lexicon.find("man", field="definition").id),
            "f6c860ea-80ff-431a-a5dd-9469b315f40c"
        )

    def test_find_gloss(self):
        self.assertEqual(
            str(self.lexicon.find("house").id),
            "086e06e3-a5f6-4c54-805a-4a1adcd51d4f"
        )

    def test_find_header_field(self):
        self.assertEqual(
            str(self.lexicon.find("1000", field="CAWL").id),
            "eee0d95e-8343-4307-83bb-944104cd2ead"
        )



"""Manipulate base linguistic elements."""

import sys
from lxml import etree
from typing import List
from typing import Optional

from . import config
from .datatypes import PCData
from .datatypes import DateTime
from .datatypes import Key
from .datatypes import Lang
from .datatypes import Prop
from .datatypes import Props
from .datatypes import URL
from .utils import etree_to_obj_attributes
from .utils import etree_to_xmlstring
from .utils import obj_attributes_to_etree


class LIFTUtilsBase:
    """This is a base class for all LIFT-related objects in this package.

    :ivar etree._Element xml_tree: The object's current data.
    """
    def __init__(self, xml_tree: etree._Element = None):
        self.xml_tree = xml_tree
        self.xml_tag = None
        self.props = Props(lift_version=config.LIFT_VERSION)
        self.props.attributes = []
        self.props.elements = []

    def print(self, _format='xml'):
        """Print the object's data to stdout; as XML by default."""
        self.xml_tree = self._to_xml_tree()
        try:
            if _format == 'xml':
                print(self._to_xml(), flush=True)
            else:
                return
        except BrokenPipeError:
            sys.stdout = None

    def _update_from_xml(self, xml_tree):
        # Set initial xml_tree.
        self.xml_tree = xml_tree
        # Update object attributes.
        etree_to_obj_attributes(xml_tree, self)

    def _to_xml_tree(self):
        xml_tree = obj_attributes_to_etree(self, self.xml_tag)
        return xml_tree

    def _to_xml(self, xml_tree=None):
        if xml_tree is None:
            xml_tree = self._to_xml_tree()
        return etree_to_xmlstring(xml_tree)


class Span(LIFTUtilsBase):
    """A Unicode string marked with language and formatting information.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('lang', prop_type=Lang),
            Prop('href', prop_type=URL),
            Prop('class_', prop_type=str),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('pcdata', required=True, prop_type=PCData),
            Prop('tail', prop_type=PCData),
            Prop('span_items', prop_type=list, item_type=Span),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'span'
        # attributes
        self.lang: Optional[Lang] = None
        self.href: Optional[URL] = None
        self.class_: Optional[str] = None
        # elements
        self.pcdata: PCData = None
        self.tail: Optional[PCData] = None
        self.span_items: Optional[List[Span]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Trait(LIFTUtilsBase):
    """An important mechanism for giving type information to an object.
    It can also be used for adding binary constraints.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('name', required=True, prop_type=Key),
            Prop('value', required=True, prop_type=Key),
            Prop('id', prop_type=Key),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        self.props.add_to(
            'elements',
            Prop('annotation_items', prop_type=list, item_type=Annotation)
        )
        self.xml_tag = 'trait'
        # attributes
        self.name: Key = None
        self.value: Key = None
        self.id: Optional[Key] = None
        # elements
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return f"{self.name}: {self.value}"


class Flag(Trait):
    """An important mechanism for giving type information to an object.
    It can also be use for adding binary constraints.

    .. note:: Used by LIFT v0.13 (FieldWorks). Mentioned in specification but
        not defined; assumed to be equivalent to ``Trait``.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()


class Text(LIFTUtilsBase):
    """Contains textual data mixed with ``span`` elements only.
    """

    def __init__(
        self,
        text: PCData = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        elems = [
            Prop('pcdata', required=True, prop_type=PCData),
            Prop('span_items', prop_type=list, item_type=Span),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'text'
        # elements
        self.pcdata = text
        self.span_items: Optional[List[Span]] = None
        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        if self.span_items:
            s = f"{str(self.pcdata)}{''.join(str(s) for s in self.span_items)}"
        else:
            s = str(self.pcdata)
        return s


class Form(LIFTUtilsBase):
    """A representation of a string in a particular language and script.
    This is specified by the ``lang`` attribute.
    """

    def __init__(
        self,
        lang: Lang = None,
        text: Text = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'attributes',
            Prop('lang', required=True, prop_type=Lang)
        )
        elems = [
            Prop('text', required=True, prop_type=Text),
            Prop('annotation_items', prop_type=list, item_type=Annotation),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'form'
        # attributes
        self.lang = lang
        # elements
        self.text = text
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return str(self.text)


class URLRef(LIFTUtilsBase):
    """This is a URL with a caption.
    """

    def __init__(
        self,
        href: URL = None,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'attributes',
            Prop('href', required=True, prop_type=URL)
        )
        self.props.add_to(
            'elements',
            Prop('label', prop_type=Multitext)
        )
        self.xml_tag = 'urlref'
        # attributes
        self.href = href
        # elements
        self.label: Optional[Multitext] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Multitext(Text):
    """Allows for different representations of the same information.
    It can be in a given language, or in multiple languages.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        elems = [
            Prop('form_items', prop_type=list, item_type=Form),
            Prop('trait_items', prop_type=list, item_type=Trait),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        # Multitext is only used as a super class for other classes, so it
        # doesn't have it's own XML tag.
        self.xml_tag = None
        # elements
        self.form_items: Optional[List[Form]] = None
        self.trait_items: Optional[List[Trait]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        s = 'multitext'
        if self.form_items:
            self.form_items.sort()
            s = f"{self.form_items[0].text} ({self.form_items[0].lang})"
            ct = len(self.form_items)
            if ct > 1:
                s = f"{s} ({ct} forms)"
        return s


class Gloss(Form):
    """A representation of a sense's gloss.
    It's given in a particular language and script as specified by the ``lang``
    attribute.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        self.props.add_to(
            'elements',
            Prop('trait_items', prop_type=list, item_type=Trait)
        )
        self.xml_tag = 'gloss'
        # elements
        self.trait_items: Optional[List[Trait]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def __str__(self):
        return f"{self.text} ({self.lang})"


class Annotation(Multitext):
    """Provides a mechanism for adding meta-information to almost any element.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('name', required=True, prop_type=Key),
            Prop('value', required=True, prop_type=Key),
            Prop('who', prop_type=Key),
            Prop('when', prop_type=DateTime),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        self.xml_tag = 'annotation'
        # attributes
        self.name: Key = None
        self.value: Key = None
        self.who: Optional[Key] = None
        self.when: Optional[DateTime] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Field(Multitext):
    """A generalised element to allow an application to store information.
    It's for information in a LIFT file that isn't explicitly described in the
    LIFT standard.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('date_created', prop_type=DateTime),
            Prop('date_modified', prop_type=DateTime),
        ]
        if config.LIFT_VERSION == '0.13':
            attribs.append(Prop('type', required=True, prop_type=Key))
        else:
            attribs.append(Prop('name', required=True, prop_type=Key))
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('annotation_items', prop_type=list, item_type=Annotation),
            Prop('trait_items', prop_type=list, item_type=Trait),
        ]
        if config.LIFT_VERSION == '0.13':
            elems.append(Prop('form_items', prop_type=list, item_type=Span))
        for e in elems:
            self.props.add_to('elements', e)
        self.xml_tag = 'field'
        # attributes
        if config.LIFT_VERSION == '0.13':
            self.type: Key = None
        else:
            self.name: Key = None
        self.date_created: Optional[DateTime] = None
        self.date_modified: Optional[DateTime] = None
        # elements
        if config.LIFT_VERSION == '0.13':
            # self.trait_items: Optional[List[Flag]] = None
            self.trait_items: Optional[List[Trait]] = None
            self.form_items: Optional[List[Span]] = None
        else:
            self.trait_items: Optional[List[Trait]] = None
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)


class Extensible(LIFTUtilsBase):
    """Used to provide certain extra information in a controlled way.

    :ivar Optional[DateTime] date_created: Contains a date/timestamp saying
        when the element was added to the dictionary.
    :ivar Optional[DateTime] date_modified: Contains a date/timestamp saying
        when the element was last changed.
    :ivar Optional[List[Field]] field_items: Holds extra textual information.
    :ivar Optional[List[Trait]] trait_items: Adds type or constraint
        information.
    :ivar Optional[List[Annotation]] annotation: Adds meta-information
        describing the element.
    """

    def __init__(
        self,
        xml_tree: Optional[etree._Element] = None
    ):
        super().__init__()
        # properties
        attribs = [
            Prop('date_created', prop_type=DateTime),
            Prop('date_modified', prop_type=DateTime),
        ]
        for a in attribs:
            self.props.add_to('attributes', a)
        elems = [
            Prop('field_items', prop_type=list, item_type=Field),
            Prop('trait_items', prop_type=list, item_type=Trait),
            Prop('annotation_items', prop_type=list, item_type=Annotation),
        ]
        for e in elems:
            self.props.add_to('elements', e)
        # Extensible is only used as a super class for other classes, so it
        # doesn't need its own XML tag.
        self.xml_tag = None
        # attributes
        self.date_created: Optional[DateTime] = None
        self.date_modified: Optional[DateTime] = None
        # elements
        self.field_items: Optional[List[Field]] = None
        if config.LIFT_VERSION == '0.13':
            # TODO: Find definition of "Flag" in v0.13?
            # self.trait_items: Optional[List[Flag]] = None
            self.trait_items: Optional[List[Trait]] = None
        else:
            self.trait_items: Optional[List[Trait]] = None
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._update_from_xml(xml_tree)

    def _add_attribs_to_xml_tree(self, xml_tree):
        for attrib in self.props.attributes:
            a = attrib.name
            x = config.XML_NAMES.get(a, a)
            xml_tree.set(x, self.__dict__.get(a))
        return xml_tree
